/* LICENSE INFO */

/*
The MIT License (MIT)

Copyright (C) 2025 Person behind the creation of the "Jatoba" account in the Macintosh Garden website (accessible from https://macintoshgarden.org/ as of April 2025)

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/* CODE */

#if defined(__MAC__) || defined(macintosh)
  #define DIRECTORY_SEPARATOR_CHARACTER ':'
#elif defined(_WIN32)
  /* Stupid crap added by Microsoft circa 2005 in an attempt
  to make C non-portable under the false premise of coding "safety",
  so we kindly tell the MS compiler to fuck off, in a way so that
  this otherwise-useless "#define" does not waste even a single
  resource from the resulting computer program on superior platforms. */
  #define _CRT_SECURE_NO_DEPRECATE

  #define DIRECTORY_SEPARATOR_CHARACTER '\\'

  /* For "_getch" function */
  #include <conio.h>
#else
  #define DIRECTORY_SEPARATOR_CHARACTER '/'
#endif

#include <stdio.h>
#include <stdlib.h>

char* getFileName(char* filePath)
{
  char* fileName = filePath;
  char* searchHelper = filePath;

  while (*searchHelper)
  {
    if (*searchHelper == DIRECTORY_SEPARATOR_CHARACTER)
    {
      fileName = searchHelper + 1;
    }

    searchHelper++;
  }

  return fileName;
}

void messageBeforeQuit(void)
{
#ifdef _WIN32
  printf("Done reading? Press any key to quit.");
  _getch();
#else
  printf("Done reading? Press Enter to quit.");
  getchar();
#endif
}

int main(int argc, char* argv[])
{
  FILE* file;
  unsigned short mz_signature, machine;
  unsigned long binarySubtypeOffset;
  unsigned char binarySubtypeSignature[4];
  unsigned char architecture[36];
  int i;

  if (argc < 2)
  {
    printf("GUI Usage: Drag-and-drop one or multiple files into %s.\n", argv[0]);
    printf("CLI Usage: %s <file_1_path> <file_2_path> ... <file_n_path>\n\n", argv[0]);
    messageBeforeQuit();

    return EXIT_FAILURE;
  }

  for (i = 1; i < argc; i++)
  {
    printf("File path: %s\n", argv[i]);
    printf("File name: %s\n", getFileName(argv[i]));
    printf("File info: ");

    file = fopen(argv[i], "rb");

    if (!file)
    {
      printf("File not found.\n\n");

      continue;
    }

    /* Check file size, even a DOS app/lib header alone is at least 28 bytes big */
    fseek(file, 0, SEEK_END);

    if (ftell(file) < 28)
    {
      printf("Too small to be a DOS, Windows or OS/2 app/lib.\n\n");
      fclose(file);

      continue;
    }

    rewind(file);

    /* Check for "MZ" (or "ZM") signature, used for DOS, Windows and OS/2 apps/libs */
    if (fread(&mz_signature, sizeof(mz_signature), 1, file) != 1 ||
        mz_signature != 0x5A4D &&
        mz_signature != 0x4D5A)
    {
      printf("Not a DOS, Windows or OS/2 app/lib (no \"MZ\" nor \"ZM\" signature).\n\n");
      fclose(file);

      continue;
    }

    rewind(file);

    /* Check file size again, the header for some DOS, and all Windows and OS/2, apps/libs are at least 64 bytes big */
    fseek(file, 0, SEEK_END);

    if (ftell(file) < 64)
    {
      printf("x86 (16-bit) DOS app/lib.\n\n");
      fclose(file);

      continue;
    }

    /* Read binary subtype offset to see if a subtype is even included */
    fseek(file, 0x3C, SEEK_SET);

    if (fread(&binarySubtypeOffset, sizeof(binarySubtypeOffset), 1, file) != 1)
    {
      printf("x86 (16-bit) DOS app/lib.\n\n");
      fclose(file);

      continue;
    }

    /* Check binary subtype signature */
    if (fseek(file, binarySubtypeOffset, SEEK_SET) != 0 ||
        fread(binarySubtypeSignature, 4, 1, file) != 1)
    {
      printf("Error when trying to check for binary subtype signature.\n\n");
      fclose(file);

      continue;
    }

    if (binarySubtypeSignature[0] == 'P' &&
        binarySubtypeSignature[1] == 'E' &&
        binarySubtypeSignature[2] == 0 &&
        binarySubtypeSignature[3] == 0)
    {
      /* "PE" (Portable Executable) Windows app/lib */
    }
    else if (binarySubtypeSignature[0] == 'N' &&
             binarySubtypeSignature[1] == 'E')
    {
      /* "NE" (New Executable) 16-bit Windows app/lib */
      printf("x86 (16-bit) Windows app/lib (\"NE\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'L' &&
             binarySubtypeSignature[1] == 'E')
    {
      /* "LE" (Linear Executable) Windows/OS/2 app/lib, generally for hardware drivers */
      printf("x86(?) (32-bit?) Windows and/or OS/2 app/lib (\"LE\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'L' &&
             binarySubtypeSignature[1] == 'X') {
      /* "LX" (Linear eXecutable) OS/2 app/lib */
      printf("x86(?) (32-bit?) Windows(?) and/or OS/2 app/lib (\"LX\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'W' &&
             binarySubtypeSignature[1] == '3')
    {
      /* "W3" Windows app/lib, e.g. Windows WIN386.EXE file: a collection of LE files */
      printf("x86 (32-bit?) Windows app/lib (\"W3\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'W' &&
             binarySubtypeSignature[1] == '4')
    {
      /* "W4" Windows app/lib, e.g. Windows95 VMM32.VXD file */
      printf("x86 (32-bit?) Windows app/lib (\"W4\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'D' &&
             binarySubtypeSignature[1] == 'L')
    {
      /* "DL" Windows app/lib, HP 100LX/200LX system manager compliant executable (.EXM) */
      printf("x86 (16-bit?) DOS app/lib (\"DL\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'M' &&
             binarySubtypeSignature[1] == 'P')
    {
      /* "MP" DOS app/lib, e.g. old PharLap .EXP */
      printf("x86 (16-bit?) DOS app/lib (\"MP\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'P' &&
             binarySubtypeSignature[1] == '2')
    {
      /* "P2" DOS app/lib, e.g. PharLap 286 .EXP */
      printf("x86 (16-bit?) DOS app/lib (\"P2\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else if (binarySubtypeSignature[0] == 'P' &&
             binarySubtypeSignature[1] == '3')
    {
      /* "P3" DOS app/lib, e.g. PharLap 386 .EXP */
      printf("x86 (16-bit?) DOS app/lib (\"P3\" subtype).\n\n");
      fclose(file);

      continue;
    }
    else
    {
      printf("x86 (16-bit) DOS app/lib.\n\n");
      fclose(file);

      continue;
    }

    /* Read machine type */
    if (fread(&machine, sizeof(machine), 1, file) != 1)
    {
      printf("Broken Windows \"PE\" app/lib: cannot read machine type.\n\n");
      fclose(file);

      continue;
    }

    /* Identify architecture */
    switch (machine)
    {
      case 0x01F0: sprintf(architecture, "PowerPC"); break; /* IMAGE_FILE_MACHINE_POWERPC: Power PC little endian */
      case 0x01F1: sprintf(architecture, "PowerPC with FPU"); break; /* IMAGE_FILE_MACHINE_POWERPCFP: Power PC with floating point support */

      case 0x0000: sprintf(architecture, "Unknown architecture"); break; /* IMAGE_FILE_MACHINE_UNKNOWN: The content of this field is assumed to be applicable to any machine type */

      case 0x0266: sprintf(architecture, "MIPS16"); break; /* IMAGE_FILE_MACHINE_MIPS16: MIPS16 */
      case 0x0466: sprintf(architecture, "MIPS16 with FPU"); break; /* IMAGE_FILE_MACHINE_MIPSFPU16: MIPS16 with FPU */
      case 0x0366: sprintf(architecture, "MIPS with FPU"); break; /* IMAGE_FILE_MACHINE_MIPSFPU: MIPS with FPU */
      case 0x0160: sprintf(architecture, "MIPS I compatible 32-bit big endian"); break; /* IMAGE_FILE_MACHINE_R3000BE: MIPS I compatible 32-bit big endian */
      case 0x0162: sprintf(architecture, "MIPS I compatible 32-bit little endian"); break; /* IMAGE_FILE_MACHINE_R3000: MIPS I compatible 32-bit little endian */
      case 0x0166: sprintf(architecture, "MIPS III compatible 64-bit little endian"); break; /* IMAGE_FILE_MACHINE_R4000: MIPS III compatible 64-bit little endian */
      case 0x0168: sprintf(architecture, "MIPS IV compatible 64-bit little endian"); break; /* IMAGE_FILE_MACHINE_R10000: MIPS IV compatible 64-bit little endian */
      case 0x0169: sprintf(architecture, "MIPS little-endian WCE v2"); break; /* IMAGE_FILE_MACHINE_WCEMIPSV2: MIPS little-endian WCE v2 */

      case 0x6232: sprintf(architecture, "LoongArch (32-bit)"); break; /* IMAGE_FILE_MACHINE_LOONGARCH32: LoongArch 32-bit processor family */
      case 0x6264: sprintf(architecture, "LoongArch (64-bit)"); break; /* IMAGE_FILE_MACHINE_LOONGARCH64: LoongArch 64-bit processor family */

      case 0x0184: sprintf(architecture, "DEC Alpha (32-bit)"); break; /* IMAGE_FILE_MACHINE_ALPHA: Alpha AXP, 32-bit address space */
      case 0x0284: sprintf(architecture, "DEC Alpha (64-bit)"); break; /* IMAGE_FILE_MACHINE_ALPHA64: Alpha 64, 64-bit address space */ /* IMAGE_FILE_MACHINE_AXP64: AXP 64 (Same as Alpha 64) */
      
      case 0x01A2: sprintf(architecture, "Hitachi SH3"); break; /* IMAGE_FILE_MACHINE_SH3: Hitachi SH3 */
      case 0x01A3: sprintf(architecture, "Hitachi SH3 DSP"); break; /* IMAGE_FILE_MACHINE_SH3DSP: Hitachi SH3 DSP */
      case 0x01A6: sprintf(architecture, "Hitachi SH4"); break; /* IMAGE_FILE_MACHINE_SH4: Hitachi SH4 */
      case 0x01A8: sprintf(architecture, "Hitachi SH5"); break; /* IMAGE_FILE_MACHINE_SH5: Hitachi SH5 */

      case 0x01D3: sprintf(architecture, "Matsushita AM33"); break; /* IMAGE_FILE_MACHINE_AM33: Matsushita AM33 */

      case 0x9041: sprintf(architecture, "Mitsubishi M32R little endian"); break; /* IMAGE_FILE_MACHINE_M32R: Mitsubishi M32R little endian */

      case 0x5032: sprintf(architecture, "RISC-V 32-bit address space"); break; /* IMAGE_FILE_MACHINE_RISCV32: RISC-V 32-bit address space */
      case 0x5064: sprintf(architecture, "RISC-V 64-bit address space"); break; /* IMAGE_FILE_MACHINE_RISCV64: RISC-V 64-bit address space */
      case 0x5128: sprintf(architecture, "RISC-V 128-bit address space"); break; /* IMAGE_FILE_MACHINE_RISCV128: RISC-V 128-bit address space */

      case 0x014C: sprintf(architecture, "x86 (32-bit)"); break; /* IMAGE_FILE_MACHINE_I386: Intel 386 or later processors and compatible processors */
      case 0x8664: sprintf(architecture, "x86 (64-bit)"); break; /* IMAGE_FILE_MACHINE_AMD64: x64 */

      case 0x0200: sprintf(architecture, "IA-64"); break; /* IMAGE_FILE_MACHINE_IA64: Intel Itanium processor family */

      case 0x01C0: sprintf(architecture, "ARM little endian"); break; /* IMAGE_FILE_MACHINE_ARM: ARM little endian */
      case 0xAA64: sprintf(architecture, "ARM64 little endian"); break; /* IMAGE_FILE_MACHINE_ARM64: ARM64 little endian */
      case 0x01C4: sprintf(architecture, "ARM Thumb-2 little endian"); break; /* IMAGE_FILE_MACHINE_ARMNT: ARM Thumb-2 little endian */

      case 0x0EBC: sprintf(architecture, "EFI byte code"); break; /* IMAGE_FILE_MACHINE_EBC: EFI byte code */

      case 0x01C2: sprintf(architecture, "Thumb"); break; /* IMAGE_FILE_MACHINE_THUMB: Thumb */

      default:     sprintf(architecture, "Undocumented (Machine ID: 0x%04X).\n", machine); break;
    }

    printf("%s Windows app/lib.\n\n", architecture);
    fclose(file);
  }

  messageBeforeQuit();

  return EXIT_SUCCESS;
}